'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var react = require('react');
var useContextSelector = require('use-context-selector');

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

var isClient = typeof window !== 'undefined' && !/ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
var useIsoLayoutEffect = isClient ? react.useLayoutEffect : react.useEffect;
var appendMap = function appendMap(dst, src) {
  src.forEach(function (v, k) {
    dst.set(k, v);
  });
  return dst;
};
var concatMap = function concatMap(src1, src2) {
  var dst = new Map();
  src1.forEach(function (v, k) {
    dst.set(k, v);
  });
  src2.forEach(function (v, k) {
    dst.set(k, v);
  });
  return dst;
};

var warningObject = new Proxy({}, {
  get: function get() {
    throw new Error('Please use <Provider>');
  },
  apply: function apply() {
    throw new Error('Please use <Provider>');
  }
});

var addDependent = function addDependent(dependentsMap, atom, dependent) {
  var dependents = dependentsMap.get(atom);

  if (!dependents) {
    dependents = new Set();
    dependentsMap.set(atom, dependents);
  }

  dependents.add(dependent);
};

var deleteDependent = function deleteDependent(dependentsMap, atom, dependent) {
  var dependents = dependentsMap.get(atom);

  if (dependents && dependents.has(dependent)) {
    dependents.delete(dependent);
    return dependents.size === 0; // empty
  }

  return false; // not found
};

var listDependents = function listDependents(dependentsMap, atom) {
  var dependents = dependentsMap.get(atom);
  return dependents || new Set();
};

var initialState = new Map();

var getAtomState = function getAtomState(state, atom) {
  var atomState = state.get(atom);

  if (!atomState) {
    throw new Error('atom is not initialized');
  }

  return atomState;
};

var getAtomStateValue = function getAtomStateValue(state, atom) {
  var atomState = state.get(atom);
  return atomState ? atomState.value : atom.initialValue;
};

var initAtom = function initAtom(id, initializingAtom, setState, dependentsMap) {
  var createAtomState = function createAtomState(prevState, atom, dependent) {
    if (dependent) {
      addDependent(dependentsMap, atom, dependent);
    }

    var partialState = new Map();
    var atomState = prevState.get(atom);

    if (atomState) {
      return partialState; // already initialized
    }

    var isSync = true;
    var nextValue = atom.read(function (a) {
      if (a !== atom) {
        if (isSync) {
          var nextPartialState = createAtomState(prevState, a, atom);
          appendMap(partialState, nextPartialState);
        } else {
          setState(function (prev) {
            var nextPartialState = createAtomState(prev, a, atom);
            return appendMap(new Map(prev), nextPartialState);
          });
        }
      }

      return getAtomStateValue(prevState, a);
    });

    if (nextValue instanceof Promise) {
      var promise = nextValue.then(function (value) {
        setState(function (prev) {
          return new Map(prev).set(atom, {
            promise: null,
            value: value
          });
        });
      });
      atomState = {
        promise: promise,
        value: atom.initialValue
      };
    } else {
      atomState = {
        promise: null,
        value: nextValue
      };
    }

    partialState.set(atom, atomState);
    isSync = false;
    return partialState;
  };

  setState(function (prev) {
    var nextPartialState = createAtomState(prev, initializingAtom, id);
    return appendMap(new Map(prev), nextPartialState);
  });
};

var disposeAtom = function disposeAtom(id, setState, dependentsMap) {
  var deleteAtomState = function deleteAtomState(prevState, dependent) {
    var nextState = new Map(prevState);
    var deleted = [];
    nextState.forEach(function (_atomState, atom) {
      var isEmpty = deleteDependent(dependentsMap, atom, dependent);

      if (isEmpty) {
        nextState.delete(atom);
        deleted.push(atom); // TODO delete in dependentsMap too (even though it is WeakMap)
      }
    });
    nextState = deleted.reduce(function (p, c) {
      return deleteAtomState(p, c);
    }, nextState);
    return nextState;
  };

  setState(function (prev) {
    return deleteAtomState(prev, id);
  });
};

var writeAtomValue = function writeAtomValue(updatingAtom, update, setState, dependentsMap, writeCache) {
  var updateDependentsState = function updateDependentsState(prevState, atom) {
    var partialState = new Map();
    listDependents(dependentsMap, atom).forEach(function (dependent) {
      if (typeof dependent === 'symbol') return;
      var v = dependent.read(function (a) {
        if (a !== dependent) {
          addDependent(dependentsMap, a, dependent);
        }

        return getAtomStateValue(prevState, a);
      });

      if (v instanceof Promise) {
        var promise = v.then(function (vv) {
          var nextAtomState = {
            promise: null,
            value: vv
          };
          setState(function (prev) {
            var nextState = new Map(prev).set(dependent, nextAtomState);
            var nextPartialState = updateDependentsState(nextState, dependent);
            return appendMap(nextState, nextPartialState);
          });
        });
        partialState.set(dependent, _extends({}, getAtomState(prevState, dependent), {
          promise: promise
        }));
      } else {
        partialState.set(dependent, {
          promise: null,
          value: v
        });
        appendMap(partialState, updateDependentsState(prevState, dependent));
      }
    });
    return partialState;
  };

  var updateAtomState = function updateAtomState(writeId, prevState, atom, value) {
    if (!writeCache.has(prevState)) {
      writeCache.set(prevState, new Map());
    }

    var cache = writeCache.get(prevState);
    var hit = cache.get(writeId);
    if (hit) return hit;
    var partialState = new Map();
    var isSync = true;
    var promise = atom.write(function (a) {
      return getAtomStateValue(concatMap(prevState, partialState), a);
    }, function (a, v) {
      if (a === atom) {
        var nextAtomState = {
          promise: null,
          value: v
        };

        if (isSync) {
          partialState.set(a, nextAtomState);
          appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), a));
        } else {
          setState(function (prev) {
            var nextState = new Map(prev).set(a, nextAtomState);
            var nextPartialState = updateDependentsState(nextState, a);
            return appendMap(nextState, nextPartialState);
          });
        }
      } else {
        var _newWriteId = Symbol();

        if (isSync) {
          var nextPartialState = updateAtomState(_newWriteId, prevState, a, v);
          appendMap(partialState, nextPartialState);
        } else {
          setState(function (prev) {
            var nextPartialState = updateAtomState(_newWriteId, prev, a, v);
            return appendMap(new Map(prev), nextPartialState);
          });
        }
      }
    }, value);

    if (promise instanceof Promise) {
      var nextAtomState = _extends({}, getAtomState(prevState, atom), {
        promise: promise.then(function () {
          setState(function (prev) {
            return new Map(prev).set(atom, _extends({}, getAtomState(prev, atom), {
              promise: null
            }));
          });
        })
      });

      partialState.set(atom, nextAtomState);
    }

    isSync = false;
    cache.set(writeId, partialState);
    return partialState;
  };

  var newWriteId = Symbol();
  setState(function (prevState) {
    var updatingAtomState = prevState.get(updatingAtom);

    if (updatingAtomState && updatingAtomState.promise) {
      // schedule update after promise is resolved
      var promise = updatingAtomState.promise.then(function () {
        var updateState = updateAtomState(newWriteId, prevState, updatingAtom, update);
        setState(function (prev) {
          return appendMap(new Map(prev), updateState);
        });
      });
      return new Map(prevState).set(updatingAtom, _extends({}, updatingAtomState, {
        promise: promise
      }));
    } else {
      var updateState = updateAtomState(newWriteId, prevState, updatingAtom, update);
      return appendMap(new Map(prevState), updateState);
    }
  });
};

var ActionsContext = useContextSelector.createContext(warningObject);
var StateContext = useContextSelector.createContext(warningObject);
var Provider = function Provider(_ref) {
  var children = _ref.children;

  var _useState = react.useState(initialState),
      state = _useState[0],
      setState = _useState[1];

  var dependentsMapRef = react.useRef();

  if (!dependentsMapRef.current) {
    dependentsMapRef.current = new WeakMap();
  }

  var writeCacheRef = react.useRef();

  if (!writeCacheRef.current) {
    writeCacheRef.current = new WeakMap();
  }

  var actions = react.useMemo(function () {
    return {
      init: function init(id, atom) {
        return initAtom(id, atom, setState, dependentsMapRef.current);
      },
      dispose: function dispose(id) {
        return disposeAtom(id, setState, dependentsMapRef.current);
      },
      write: function write(atom, update) {
        return writeAtomValue(atom, update, setState, dependentsMapRef.current, writeCacheRef.current);
      }
    };
  }, []);
  return react.createElement(ActionsContext.Provider, {
    value: actions
  }, react.createElement(StateContext.Provider, {
    value: state
  }, children));
};

function atom(read, write) {
  var instance = {};

  if (typeof read === 'function') {
    instance.read = read;
    var value = read(function (a) {
      return a.initialValue;
    });

    if (value instanceof Promise) {
      value.then(function (v) {
        instance.initialValue = v;
      });
    }

    instance.initialValue = value;
  } else {
    instance.initialValue = read;

    instance.read = function (get) {
      return get(instance);
    };

    instance.write = function (get, set, update) {
      set(instance, typeof update === 'function' ? update(get(instance)) : update);
    };
  }

  if (write) {
    instance.write = write;
  }

  return instance;
}

var isWritable = function isWritable(atom) {
  return !!atom.write;
};

function useAtom(atom) {
  var actions = useContextSelector.useContext(ActionsContext);
  var promiseOrValue = useContextSelector.useContextSelector(StateContext, react.useCallback(function (state) {
    var atomState = state.get(atom);

    if (atomState) {
      return atomState.promise || atomState.value;
    }

    if (atom.initialValue instanceof Promise) {
      atom.initialValue.then(function () {
        actions.init(null, atom);
      });
    }

    return atom.initialValue;
  }, [atom]));
  var setAtom = react.useCallback(function (update) {
    if (isWritable(atom)) {
      actions.write(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [atom, actions]);
  useIsoLayoutEffect(function () {
    var id = Symbol();
    actions.init(id, atom);
    return function () {
      actions.dispose(id);
    };
  }, [actions, atom]);

  if (promiseOrValue instanceof Promise) {
    throw promiseOrValue;
  }

  react.useDebugValue(promiseOrValue);
  return [promiseOrValue, setAtom];
}

exports.Provider = Provider;
exports.atom = atom;
exports.useAtom = useAtom;
