import { useLayoutEffect, useEffect, useState, useRef, useMemo, createElement, useCallback, useDebugValue } from 'react';
import { createContext, useContext, useContextSelector } from 'use-context-selector';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

const isClient = typeof window !== 'undefined' && !/ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
const useIsoLayoutEffect = isClient ? useLayoutEffect : useEffect;
const appendMap = (dst, src) => {
  src.forEach((v, k) => {
    dst.set(k, v);
  });
  return dst;
};
const concatMap = (src1, src2) => {
  const dst = new Map();
  src1.forEach((v, k) => {
    dst.set(k, v);
  });
  src2.forEach((v, k) => {
    dst.set(k, v);
  });
  return dst;
};

const warningObject = new Proxy({}, {
  get() {
    throw new Error('Please use <Provider>');
  },

  apply() {
    throw new Error('Please use <Provider>');
  }

});

const addDependent = (dependentsMap, atom, dependent) => {
  let dependents = dependentsMap.get(atom);

  if (!dependents) {
    dependents = new Set();
    dependentsMap.set(atom, dependents);
  }

  dependents.add(dependent);
};

const deleteDependent = (dependentsMap, atom, dependent) => {
  const dependents = dependentsMap.get(atom);

  if (dependents && dependents.has(dependent)) {
    dependents.delete(dependent);
    return dependents.size === 0; // empty
  }

  return false; // not found
};

const listDependents = (dependentsMap, atom) => {
  const dependents = dependentsMap.get(atom);
  return dependents || new Set();
};

const initialState = new Map();

const getAtomState = (state, atom) => {
  const atomState = state.get(atom);

  if (!atomState) {
    throw new Error('atom is not initialized');
  }

  return atomState;
};

const getAtomStateValue = (state, atom) => {
  const atomState = state.get(atom);
  return atomState ? atomState.value : atom.initialValue;
};

const initAtom = (id, initializingAtom, setState, dependentsMap) => {
  const createAtomState = (prevState, atom, dependent) => {
    if (dependent) {
      addDependent(dependentsMap, atom, dependent);
    }

    const partialState = new Map();
    let atomState = prevState.get(atom);

    if (atomState) {
      return partialState; // already initialized
    }

    let isSync = true;
    const nextValue = atom.read(a => {
      if (a !== atom) {
        if (isSync) {
          const nextPartialState = createAtomState(prevState, a, atom);
          appendMap(partialState, nextPartialState);
        } else {
          setState(prev => {
            const nextPartialState = createAtomState(prev, a, atom);
            return appendMap(new Map(prev), nextPartialState);
          });
        }
      }

      return getAtomStateValue(prevState, a);
    });

    if (nextValue instanceof Promise) {
      const promise = nextValue.then(value => {
        setState(prev => new Map(prev).set(atom, {
          promise: null,
          value
        }));
      });
      atomState = {
        promise,
        value: atom.initialValue
      };
    } else {
      atomState = {
        promise: null,
        value: nextValue
      };
    }

    partialState.set(atom, atomState);
    isSync = false;
    return partialState;
  };

  setState(prev => {
    const nextPartialState = createAtomState(prev, initializingAtom, id);
    return appendMap(new Map(prev), nextPartialState);
  });
};

const disposeAtom = (id, setState, dependentsMap) => {
  const deleteAtomState = (prevState, dependent) => {
    let nextState = new Map(prevState);
    const deleted = [];
    nextState.forEach((_atomState, atom) => {
      const isEmpty = deleteDependent(dependentsMap, atom, dependent);

      if (isEmpty) {
        nextState.delete(atom);
        deleted.push(atom); // TODO delete in dependentsMap too (even though it is WeakMap)
      }
    });
    nextState = deleted.reduce((p, c) => deleteAtomState(p, c), nextState);
    return nextState;
  };

  setState(prev => deleteAtomState(prev, id));
};

const writeAtomValue = (updatingAtom, update, setState, dependentsMap, writeCache) => {
  const updateDependentsState = (prevState, atom) => {
    const partialState = new Map();
    listDependents(dependentsMap, atom).forEach(dependent => {
      if (typeof dependent === 'symbol') return;
      const v = dependent.read(a => {
        if (a !== dependent) {
          addDependent(dependentsMap, a, dependent);
        }

        return getAtomStateValue(prevState, a);
      });

      if (v instanceof Promise) {
        const promise = v.then(vv => {
          const nextAtomState = {
            promise: null,
            value: vv
          };
          setState(prev => {
            const nextState = new Map(prev).set(dependent, nextAtomState);
            const nextPartialState = updateDependentsState(nextState, dependent);
            return appendMap(nextState, nextPartialState);
          });
        });
        partialState.set(dependent, _extends({}, getAtomState(prevState, dependent), {
          promise
        }));
      } else {
        partialState.set(dependent, {
          promise: null,
          value: v
        });
        appendMap(partialState, updateDependentsState(prevState, dependent));
      }
    });
    return partialState;
  };

  const updateAtomState = (writeId, prevState, atom, value) => {
    if (!writeCache.has(prevState)) {
      writeCache.set(prevState, new Map());
    }

    const cache = writeCache.get(prevState);
    const hit = cache.get(writeId);
    if (hit) return hit;
    const partialState = new Map();
    let isSync = true;
    const promise = atom.write(a => getAtomStateValue(concatMap(prevState, partialState), a), (a, v) => {
      if (a === atom) {
        const nextAtomState = {
          promise: null,
          value: v
        };

        if (isSync) {
          partialState.set(a, nextAtomState);
          appendMap(partialState, updateDependentsState(concatMap(prevState, partialState), a));
        } else {
          setState(prev => {
            const nextState = new Map(prev).set(a, nextAtomState);
            const nextPartialState = updateDependentsState(nextState, a);
            return appendMap(nextState, nextPartialState);
          });
        }
      } else {
        const newWriteId = Symbol();

        if (isSync) {
          const nextPartialState = updateAtomState(newWriteId, prevState, a, v);
          appendMap(partialState, nextPartialState);
        } else {
          setState(prev => {
            const nextPartialState = updateAtomState(newWriteId, prev, a, v);
            return appendMap(new Map(prev), nextPartialState);
          });
        }
      }
    }, value);

    if (promise instanceof Promise) {
      const nextAtomState = _extends({}, getAtomState(prevState, atom), {
        promise: promise.then(() => {
          setState(prev => new Map(prev).set(atom, _extends({}, getAtomState(prev, atom), {
            promise: null
          })));
        })
      });

      partialState.set(atom, nextAtomState);
    }

    isSync = false;
    cache.set(writeId, partialState);
    return partialState;
  };

  const newWriteId = Symbol();
  setState(prevState => {
    const updatingAtomState = prevState.get(updatingAtom);

    if (updatingAtomState && updatingAtomState.promise) {
      // schedule update after promise is resolved
      const promise = updatingAtomState.promise.then(() => {
        const updateState = updateAtomState(newWriteId, prevState, updatingAtom, update);
        setState(prev => appendMap(new Map(prev), updateState));
      });
      return new Map(prevState).set(updatingAtom, _extends({}, updatingAtomState, {
        promise
      }));
    } else {
      const updateState = updateAtomState(newWriteId, prevState, updatingAtom, update);
      return appendMap(new Map(prevState), updateState);
    }
  });
};

const ActionsContext = createContext(warningObject);
const StateContext = createContext(warningObject);
const Provider = ({
  children
}) => {
  const [state, setState] = useState(initialState);
  const dependentsMapRef = useRef();

  if (!dependentsMapRef.current) {
    dependentsMapRef.current = new WeakMap();
  }

  const writeCacheRef = useRef();

  if (!writeCacheRef.current) {
    writeCacheRef.current = new WeakMap();
  }

  const actions = useMemo(() => ({
    init: (id, atom) => initAtom(id, atom, setState, dependentsMapRef.current),
    dispose: id => disposeAtom(id, setState, dependentsMapRef.current),
    write: (atom, update) => writeAtomValue(atom, update, setState, dependentsMapRef.current, writeCacheRef.current)
  }), []);
  return createElement(ActionsContext.Provider, {
    value: actions
  }, createElement(StateContext.Provider, {
    value: state
  }, children));
};

function atom(read, write) {
  const instance = {};

  if (typeof read === 'function') {
    instance.read = read;
    const value = read(a => a.initialValue);

    if (value instanceof Promise) {
      value.then(v => {
        instance.initialValue = v;
      });
    }

    instance.initialValue = value;
  } else {
    instance.initialValue = read;

    instance.read = get => get(instance);

    instance.write = (get, set, update) => {
      set(instance, typeof update === 'function' ? update(get(instance)) : update);
    };
  }

  if (write) {
    instance.write = write;
  }

  return instance;
}

const isWritable = atom => !!atom.write;

function useAtom(atom) {
  const actions = useContext(ActionsContext);
  const promiseOrValue = useContextSelector(StateContext, useCallback(state => {
    const atomState = state.get(atom);

    if (atomState) {
      return atomState.promise || atomState.value;
    }

    if (atom.initialValue instanceof Promise) {
      atom.initialValue.then(() => {
        actions.init(null, atom);
      });
    }

    return atom.initialValue;
  }, [atom]));
  const setAtom = useCallback(update => {
    if (isWritable(atom)) {
      actions.write(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [atom, actions]);
  useIsoLayoutEffect(() => {
    const id = Symbol();
    actions.init(id, atom);
    return () => {
      actions.dispose(id);
    };
  }, [actions, atom]);

  if (promiseOrValue instanceof Promise) {
    throw promiseOrValue;
  }

  useDebugValue(promiseOrValue);
  return [promiseOrValue, setAtom];
}

export { Provider, atom, useAtom };
